import numpy as np
import matplotlib.pyplot as plt

def init_ax_equal_axis(ax, xlim, ylim):
    ax.set_aspect('equal', adjustable='box')
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)


def draw_vector_2d(ax, vec, origin, color='black'):
    ax.arrow(origin[0], origin[1], vec[0], vec[1], head_width=0.1, head_length=0.1, color=color)

def draw_coord_sys_2d(ax, x_vec, y_vec, orig, size=1.0, label="", color=None):
    if color is None:
        x_color = 'red'
        y_color = 'green'
    else:
        x_color = color
        y_color = color
    draw_vector_2d(ax, x_vec, orig, color=x_color)
    draw_vector_2d(ax, y_vec, orig, color=y_color)
    text_pos = np.array(orig)-0.25*(np.array(x_vec)+np.array(y_vec))-0.2
    plt.text(text_pos[0],text_pos[1],label)

def draw_arrowhead(ax, vec, origin, color='black', label=""):
    label_pos = np.array(origin) - 0.35*np.array([vec[1], vec[0]])-0.2
    ax.arrow(origin[0], origin[1], vec[0]*0.0001, vec[1]*0.0001, head_width=0.3, head_length=0.5, edgecolor=color, facecolor='white')
    plt.text(label_pos[0],label_pos[1],label)






if __name__ == '__main__':
    fig,ax = plt.subplots()
    vec = [1,0]
    draw_vector_2d(ax, vec)
    plt.show()





import numpy as np
import math
import matplotlib.pyplot as plt
from rhovee import manifolds
    
def solve_RKMK_row(A_row, Fs):
    Theta = np.zeros(Fs[0].shape[0])
    for j in range(len(Fs)):
        Theta+=Fs[j]*A_row[j]
    return Theta

def sum_to_current_est(Fs, Bs):
    assert (len(Fs) == len(Bs))
    sum_vec = np.zeros(Fs[0].shape[0])
    for i in range(len(Fs)):
        sum_vec += Bs[i]*Fs[i]
    return sum_vec

class ButcherTableu():
    @property
    def heun(self):
        As = np.array([[0.0,0.0],[1.0,0.0]])
        Bs = np.array([0.5,0.5])
        Cs = np.array([0.0,1.0])
        heuns = (As,Bs,Cs)
        return heuns

    @property
    def euler(self):
        As = np.array([[0.0]])
        Bs = np.array([1.0])
        Cs = np.array([0.0])
        return (As,Bs,Cs)

    @property
    def rk4(self):
        As = np.zeros((4,4))
        As[1,0] = 0.5
        As[2,1] = 0.5
        As[3,2] = 1.0
        Bs = np.array([1.0/6.0, 1.0/3.0, 1.0/3.0, 1.0/6.0])
        Cs = np.array([0.0, 0.5,0.5,1.0])
        return (As,Bs,Cs)

butcher_tableu = ButcherTableu()



def RKMK(steps, kinematics, sim_duration, butcher_tableu, manifold):
    assert isinstance(butcher_tableu, tuple)
    mf = manifold
    mf_shape = mf.manifold_dim()
    mf_vec_shape = mf.vector_dim()
    As,Bs,Cs = butcher_tableu
    A_rows, A_cols = As.shape
    assert A_rows == len(Cs)
    assert A_cols == len(Bs)
    result = []

    stages = len(Cs)
    time = 0
    h = sim_duration/steps
    T0 = np.identity(mf_shape[0])

    T = T0
    result.append(T)
    for step in range(0,steps):
        Fs = []
        Theta0 = np.zeros(mf_vec_shape[0])
        F0 = h*kinematics(T0, time+Cs[0]*h)
        Fs.append(F0)
        for s in range(1,stages):
            A_row = As[s,:]
            Theta = solve_RKMK_row(A_row, Fs)
            theta = h*kinematics(T@mf.Exp(Theta), time+Cs[s]*h)
            inv_jac = mf.inv_right_jacobian(Theta)
            F = inv_jac@theta
            Fs.append(F)
        Theta = sum_to_current_est(Fs, Bs)
        T = T@mf.Exp(Theta)
        result.append(T)
        time += h
    return result
            



import numpy as np
import cv2


def draw_centered_line(img_size, angle, line_width):
    # draw a line that goes through the center of the image
    img = np.zeros(img_size)
    center = np.array(img_size) / 2
    x1 = 0
    y1 = int(center[1] - np.tan(angle) * center[0])
    x2 = img.shape[1]
    y2 = int(center[1] + np.tan(angle) * (img.shape[1] - center[0]))
    cv2.line(img, (x1, y1), (x2, y2), (255, 0, 0), 5)
    return img

def weighted_mean_row_index(img, threshold=20, verbose=0):
    assert img.ndim == 2
    img = np.where(img > threshold, img, 0)
    if verbose:
        cv2.imshow('img', img)
        cv2.waitKey(0)
    # find index weighted mean
    indeces = np.arange(img.shape[1])
    mean = np.sum(indeces * img, axis=1) / np.sum(img, axis=1)
    return mean

def row_list_to_points(row_list):
    ys = np.arange(len(row_list))
    xs = row_list
    return np.vstack([xs, ys, np.ones(len(xs))]).T

def draw_homg_line(img, l):
    assert l.shape[0] == 3
    out = img.copy()
    x1 = 0
    y1 = int(-l[2] / l[1])
    x2 = img.shape[1]
    y2 = int(-(l[2] + l[0] * x2) / l[1])
    cv2.line(out, (x1, y1), (x2, y2), (255, 0, 0), 1)
    return out

def fit_homogenous_line(points):
   assert points.shape[1] == 3
   # find m,c using least squares
   points = np.asarray(points)
   n, _ = points.shape
   mean_x = np.mean(points[:,0])
   mean_y = np.mean(points[:,1])
   sum_xy = sum([points[i,0]*points[i,1] for i in range(n)])
   sum_x = sum([points[i,0] for i in range(n)])
   sum_y = sum([points[i,1] for i in range(n)])
   sum_x2 = sum([points[i,0]**2 for i in range(n)])
   slope = (n*sum_xy - sum_x*sum_y) / (n*sum_x2 - sum_x**2)
   intercept = mean_y - slope*mean_x
   return np.array([-slope, 1, -intercept])

def overlap_gray_imgs_rgb(img1, img2):
    assert img1.shape == img2.shape
    rgb_overlap = np.dstack([img1, img2, np.zeros(img1.shape)])
    return rgb_overlap




if __name__ == '__main__':
    img = draw_centered_line((500, 500), np.pi / 2.1, 5)
    # blur img
    img = cv2.GaussianBlur(img, (5, 5), 0)
    # add noise with max value 10
    img = img + np.random.randint(0, 10, img.shape)
    row_means = weighted_mean_row_index(img, verbose=1, threshold=50)
    mean_points = row_list_to_points(row_means)
    l = fit_homogenous_line(mean_points)
    zero_img = np.zeros(img.shape)
    out = draw_homg_line(zero_img, l)
    rgb_overlap = overlap_gray_imgs_rgb(img, out)
    cv2.imshow('rgb_overlap', rgb_overlap)
    cv2.waitKey(0)






def pc_to_o3d(pc_np):
    assert pc_np.shape[1] == 3
    


def o3d_to_pc(pc_o3d):
    return np.array(pc_o3d.points)

def keep_overlap_pc(pc1_np, pc2_np, max_dist):
    # pc shape (n,3)
    assert pc1_np.shape[1] == 3 and pc2_np.shape[1] == 3
    pc1_o3d = o3d.geometry.PointCloud()
    pc2_o3d = o3d.geometry.PointCloud()
    pc1_o3d.points = o3d.utility.Vector3dVector(pc1_np)
    pc2_o3d.points = o3d.utility.Vector3dVector(pc2_np)
    result = o3d.pipelines.registration.evaluate_registration(pc1_o3d, pc2_o3d, max_dist, np.eye(4))
    corr = np.array(result.correspondence_set)
    kept_inds = corr[:,0]
    pc1_points = np.asarray(pc1_o3d.points)
    kept_points = pc1_points[pc1_keep_inds]
    return kept_points
    


    



if __name__ == '__main__':
    pass
